var React = require('react');
var slate = require('slate');
var slateHistory = require('slate-history');
var slateReact = require('slate-react');
var ulid = require('ulid');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var React__default = /*#__PURE__*/_interopDefaultLegacy(React);

function _extends() {
  _extends = Object.assign ? Object.assign.bind() : function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];
      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }
    return target;
  };
  return _extends.apply(this, arguments);
}

function debounce(func, waitMilliseconds, options) {
  var _options$isImmediate, _options$callback;
  if (waitMilliseconds === void 0) {
    waitMilliseconds = 50;
  }
  if (options === void 0) {
    options = {};
  }
  var timeoutId;
  var isImmediate = (_options$isImmediate = options.isImmediate) != null ? _options$isImmediate : false;
  var callback = (_options$callback = options.callback) != null ? _options$callback : false;
  var maxWait = options.maxWait;
  var lastInvokeTime = Date.now();
  var promises = [];
  function nextInvokeTimeout() {
    if (maxWait !== undefined) {
      var timeSinceLastInvocation = Date.now() - lastInvokeTime;
      if (timeSinceLastInvocation + waitMilliseconds >= maxWait) {
        return maxWait - timeSinceLastInvocation;
      }
    }
    return waitMilliseconds;
  }
  var debouncedFunction = function debouncedFunction() {
    var args = [].slice.call(arguments);
    var context = this;
    return new Promise(function (resolve, reject) {
      var invokeFunction = function invokeFunction() {
        timeoutId = undefined;
        lastInvokeTime = Date.now();
        if (!isImmediate) {
          var result = func.apply(context, args);
          callback && callback(result);
          promises.forEach(function (_ref) {
            var resolve = _ref.resolve;
            return resolve(result);
          });
          promises = [];
        }
      };
      var shouldCallNow = isImmediate && timeoutId === undefined;
      if (timeoutId !== undefined) {
        clearTimeout(timeoutId);
      }
      timeoutId = setTimeout(invokeFunction, nextInvokeTimeout());
      if (shouldCallNow) {
        var result = func.apply(context, args);
        callback && callback(result);
        return resolve(result);
      }
      promises.push({
        resolve: resolve,
        reject: reject
      });
    });
  };
  debouncedFunction.cancel = function (reason) {
    if (timeoutId !== undefined) {
      clearTimeout(timeoutId);
    }
    promises.forEach(function (_ref2) {
      var reject = _ref2.reject;
      return reject(reason);
    });
    promises = [];
  };
  return debouncedFunction;
}

/**
 * Returns a random integer between min (inclusive) and max (inclusive).
 * The value is no lower than min (or the next integer greater than min
 * if min isn't an integer) and no greater than max (or the next integer
 * lower than max if max isn't an integer).
 * Using Math.round() will give you a non-uniform distribution!
 */
function getRandomInt(min, max) {
  min = Math.ceil(min);
  max = Math.floor(max);
  return Math.floor(Math.random() * (max - min + 1)) + min;
}

function throttle(func, limit) {
  var inThrottle;
  var lastResult;
  return function () {
    var args = arguments;
    var context = this;
    if (!inThrottle) {
      inThrottle = true;
      setTimeout(function () {
        return inThrottle = false;
      }, limit);
      lastResult = func.apply(context, args);
    }
    return lastResult;
  };
}

/**
 * Responsible for fetching completions from a model and adding them to the input.
 * Each `<SmartInput />` component rendered to the DOM will have its own instance of this class as state.
 */
var SmartInputFetcher = /*#__PURE__*/function () {
  /**
   * sequenceId is used to ensure that we don't add completions from a previous fetch,
   * we increment it each time we request to fetch a new completion because the user input has changed
   */

  /**
   * sessionId is used to track this particular instantiation of the `<SmartInput />` component
   */

  /**
   * completionCache is used to cache the completions that we've already fetched from the model
   */

  /**
   * alreadySeen tracks which completions we've marked as seen (reduce chattiness)
   */

  /**
   * alreadyAccepted tracks which completions we've marked as acceoted (reduce chattiness)
   */

  /**
   * currentCompletion tracks the latest text that was sent to the model for fetching completions, and the
   * completion that was returned (if any ~ we may still be waiting for it)
   */

  function SmartInputFetcher(apiUrl, addCompletion, modelId, mockCompletion) {
    var _this2 = this;
    var _this = this;
    this.apiUrl = void 0;
    this.addCompletion = void 0;
    this.modelId = void 0;
    this.mockCompletion = void 0;
    this.sequenceId = 0;
    this.sessionId = ulid.ulid();
    this.completionCache = {};
    this.alreadySeen = new Set();
    this.alreadyAccepted = new Set();
    this.currentCompletion = null;
    this.fetchCompletion = debounce(function (text, context) {
      try {
        var _this$currentCompleti2;
        var _temp4 = function _temp4(_result2) {
          return _exit ? _result2 : Promise.resolve(fetch(_this.apiUrl + "/completions", {
            method: 'POST',
            mode: 'cors',
            headers: {
              'Content-Type': 'application/json',
              Authorization: "Bearer " + _this.modelId
            },
            body: JSON.stringify({
              model: _this.modelId,
              input: text,
              session: _this.sessionId,
              context: context
            })
          })).then(function (res) {
            if (res.status !== 200) {
              console.error("Couldn't fetch completions");
              return null;
            }
            return Promise.resolve(res.json()).then(function (data) {
              var _this$currentCompleti;
              _this.completionCache[text] = data;
              if (((_this$currentCompleti = _this.currentCompletion) == null ? void 0 : _this$currentCompleti.sequenceId) === seqId) {
                _this.addCompletion(text, data);
                _this.currentCompletion.completion = data.completion;
              }
              return data;
            });
          });
        };
        var _exit;
        if (text.length === 0) {
          // only get completions for non-empty text
          return Promise.resolve(null);
        }
        var seqId = ++_this.sequenceId;
        // if a completion was already added, and the user is typing the same characters as the current completion, don't fetch
        if ((_this$currentCompleti2 = _this.currentCompletion) != null && _this$currentCompleti2.completion) {
          var currExpectedText = _this.currentCompletion.text + _this.currentCompletion.completion;
          if (currExpectedText.startsWith(text)) {
            return Promise.resolve(null);
          }
        }
        // update the current completion
        _this.currentCompletion = {
          text: text,
          sequenceId: seqId
        };
        // if we already have a completion for this text, don't fetch, but do return it
        var cachedVal = _this.completionCache[text];
        if (cachedVal) {
          // don't fetch the same text twice
          _this.addCompletion(text, cachedVal);
          _this.currentCompletion.completion = cachedVal.completion;
          return Promise.resolve(cachedVal);
        }
        var _temp3 = function () {
          if (_this.mockCompletion) {
            var _temp2 = function _temp2(_result) {
              if (_exit) return _result;
              var _input$sid$completion = {
                input: text,
                sid: ulid.ulid(),
                completion: null
              };
              _exit = 1;
              return _input$sid$completion;
            };
            var completion = _this.mockCompletion(text);
            var _temp = function () {
              if (completion) {
                return Promise.resolve(new Promise(function (resolve) {
                  setTimeout(function () {
                    var _this$currentCompleti3;
                    var c = {
                      input: text,
                      sid: ulid.ulid(),
                      completion: completion
                    };
                    _this.completionCache[text] = c;
                    if (((_this$currentCompleti3 = _this.currentCompletion) == null ? void 0 : _this$currentCompleti3.sequenceId) === seqId) {
                      _this.addCompletion(text, c);
                      _this.currentCompletion.completion = completion;
                    }
                    resolve(c);
                  }, getRandomInt(200, 1000)); // simulate API latency
                })).then(function (_await$Promise) {
                  _exit = 1;
                  return _await$Promise;
                });
              }
            }();
            return _temp && _temp.then ? _temp.then(_temp2) : _temp2(_temp);
          }
        }();
        return Promise.resolve(_temp3 && _temp3.then ? _temp3.then(_temp4) : _temp4(_temp3));
      } catch (e) {
        return Promise.reject(e);
      }
    }, 200);
    this.sendEvents = throttle(function (text, context) {
      fetch(_this2.apiUrl + "/events", {
        method: 'POST',
        mode: 'cors',
        headers: {
          'Content-Type': 'application/json',
          Authorization: "Bearer " + _this2.modelId
        },
        body: JSON.stringify({
          model: _this2.modelId,
          session: _this2.sessionId,
          input: text,
          context: context
        })
      }).then(function (res) {
        if (res.status !== 200) {
          console.error('Error sending events');
        }
      })["catch"](function (e) {
        return console.error(e);
      });
    }, 1000);
    this.apiUrl = apiUrl;
    this.addCompletion = addCompletion;
    this.modelId = modelId;
    this.mockCompletion = mockCompletion;
  }
  var _proto = SmartInputFetcher.prototype;
  _proto.updateCompletion = function updateCompletion(completionSid, arg) {
    var _this3 = this;
    if (this.mockCompletion) {
      return;
    }
    // reduce chattiness
    var shouldUpdate = arg.seen && !this.alreadySeen.has(completionSid) || arg.accepted && !this.alreadyAccepted.has(completionSid);
    if (!shouldUpdate) {
      return;
    }
    fetch(this.apiUrl + "/completions", {
      method: 'PUT',
      mode: 'cors',
      headers: {
        'Content-Type': 'application/json',
        Authorization: "Bearer " + this.modelId
      },
      body: JSON.stringify({
        completionSid: completionSid,
        seen: arg.seen,
        accepted: arg.accepted
      })
    }).then(function (res) {
      if (res.status !== 200) {
        console.error("Couldn't update completion");
      }
      if (arg.seen) {
        // only update the set if the update was successful
        _this3.alreadySeen.add(completionSid);
      }
      if (arg.accepted) {
        // only update the set if the update was successful
        _this3.alreadyAccepted.add(completionSid);
      }
    })["catch"](function (e) {
      return console.error(e);
    });
  };
  return SmartInputFetcher;
}();

console.log('here yep yep');
/**
 * A text input that provides intelligent, tab-completable completions based on the text that has already been entered.
 * Due to the limitations of `<input>` and `<textarea>` elements, this component renders a `<div contenteditable="true">` instead.
 */
function SmartInput(props) {
  var editor = React__default["default"].useMemo(function () {
    return slateHistory.withHistory(slateReact.withReact(slate.createEditor()));
  }, []);
  var _React$useState = React__default["default"].useState(null),
    completion = _React$useState[0],
    setCompletion = _React$useState[1];
  var _React$useState2 = React__default["default"].useState(new SmartInputFetcher(props.devApiUrl || 'https://api.foundation-ui.com/smartinput', function (text, completion) {
      addCompletion(completion);
    }, props.modelId, props.modelId ? props.mockCompletion : function () {
      return ' -- sample completion --';
    })),
    fetcher = _React$useState2[0];
  var acceptCompletion = function acceptCompletion() {
    EditorHelper.acceptCompletion(editor, fetcher, completion);
    setCompletion(null);
  };
  var addCompletion = function addCompletion(completion) {
    if (props.value) {
      // don't add completion if the text is gone
      EditorHelper.addCompletion(editor, fetcher, completion);
      setCompletion(completion);
    }
  };
  var clearCompletion = function clearCompletion() {
    EditorHelper.removeCompletion(editor);
    setCompletion(null);
  };
  // for good measure, always overwrite the fetcher fields
  // fetcher.modelId = props.modelId
  fetcher.addCompletion = function (text, completion) {
    var currText = props.value;
    if (completion && (text + completion).startsWith(currText)) {
      addCompletion(completion);
    }
  };
  // fetcher.mockCompletion = props.mockCompletion
  var initialValue = [{
    type: 'text',
    children: [{
      text: props.value,
      type: undefined
    }]
  }];
  React.useEffect(function () {
    if (props.editorRef) {
      props.editorRef.current = _extends({}, slateReact.ReactEditor.toDOMNode(editor, editor), {
        focus: function focus() {
          setTimeout(function () {
            slateReact.ReactEditor.focus(editor);
            if (props.value) {
              slate.Transforms.select(editor, slate.Editor.end(editor, []));
            } else {
              slate.Transforms.select(editor, slate.Editor.start(editor, []));
            }
          }, 50);
        }
      });
    }
  });
  var renderElement = React__default["default"].useCallback(function (p) {
    switch (p.element.type) {
      case 'text':
        return props.renderText(_extends({}, p));
      case 'completion':
        return props.renderCompletion(_extends({}, p, {
          contentEditable: 'false',
          suppressContentEditableWarning: true
        }));
      default:
        return /*#__PURE__*/React__default["default"].createElement(slateReact.DefaultElement, p);
    }
  },
  // eslint-disable-next-line react-hooks/exhaustive-deps
  [props.renderText, props.renderCompletion]);
  var renderPlaceholder = React__default["default"].useCallback(function (p) {
    return props.renderPlaceholder(_extends({}, p, p.attributes || {}, {
      style: _extends({
        // these style attributes usually come from slate, we gotta make sure to add them back or the cursor might go after the placeholder lol
        position: 'absolute',
        pointerEvents: 'none',
        width: '100%',
        maxWidth: '100%',
        userSelect: 'none',
        textDecoration: 'none'
      }, p.style || {})
    }));
  }, [props.renderPlaceholder]);
  return /*#__PURE__*/React__default["default"].createElement(slateReact.Slate, {
    editor: editor,
    value: initialValue,
    onChange: function onChange(value) {
      var _fetcher$fetchComplet;
      var text = EditorHelper.serialize(value.filter(function (n) {
        return n.type !== 'completion';
      }));
      (_fetcher$fetchComplet = fetcher.fetchCompletion(text, props.context || {})) == null ? void 0 : _fetcher$fetchComplet["catch"](function (e) {
        return console.error(e);
      });
      fetcher.sendEvents(text, props.context || {});
      props.onChange(text);
    }
  }, /*#__PURE__*/React__default["default"].createElement(slateReact.Editable, {
    placeholder: props.placeholder,
    className: props.className,
    style: _extends({}, props.style || {}, {
      whiteSpace: props.multiline ? undefined : 'pre'
    }),
    onPaste: function onPaste(event) {
      clearCompletion();
      event.preventDefault();
      var value = event.clipboardData.getData('Text');
      if (!props.multiline) {
        editor.insertText(value.replaceAll('\n', ''));
      } else {
        editor.insertText(value);
      }
    },
    renderPlaceholder: props.renderPlaceholder ? renderPlaceholder : undefined,
    renderElement: renderElement,
    onBlur: function onBlur(e) {
      clearCompletion();
      fetcher.currentCompletion = null;
      if (props.onBlur) {
        props.onBlur(e);
      }
    },
    onFocus: function onFocus(e) {
      if (props.onFocus) {
        props.onFocus(e);
        if (editor.selection == null) {
          // fix bug for chrome
          if (props.value) {
            slate.Transforms.select(editor, slate.Editor.end(editor, []));
          } else {
            slate.Transforms.select(editor, slate.Editor.start(editor, []));
          }
        }
      }
    },
    onKeyDown: function onKeyDown(event) {
      var _completion$completio, _completion$completio2;
      var hasCompletion = Boolean(completion == null ? void 0 : (_completion$completio = completion.completion) == null ? void 0 : (_completion$completio2 = _completion$completio.trim()) == null ? void 0 : _completion$completio2.length);
      if (event.key === 'Tab' && hasCompletion) {
        event.preventDefault();
        acceptCompletion();
        fetcher.currentCompletion = null;
      } else if (event.key === 'ArrowRight' && hasCompletion) {
        event.preventDefault();
        acceptCompletion();
        fetcher.currentCompletion = null;
      } else if (event.key === 'ArrowLeft' && hasCompletion) {
        clearCompletion();
        fetcher.currentCompletion = null;
      } else if ((event.key === 'ArrowDown' || event.key === 'ArrowUp') && hasCompletion) {
        clearCompletion();
      } else if (event.key === 'Backspace') {
        fetcher.currentCompletion = null;
        clearCompletion();
      } else if (event.key === 'Enter' && !event.metaKey) {
        event.preventDefault();
        if (props.multiline) {
          clearCompletion();
          editor.insertText('\n');
        }
      } else if (event.key !== 'Shift') {
        var _completion$completio3;
        if (completion && !event.metaKey && (_completion$completio3 = completion.completion) != null && _completion$completio3.startsWith(event.key) && event.key.length === 1) {
          var newCompletionText = completion.completion.substring(1);
          // don't clear the completion if the user is just typing the completion character for character
          EditorHelper.updateCompletion(editor, newCompletionText);
          setCompletion(_extends({}, completion, {
            completion: newCompletionText
          }));
        } else {
          fetcher.currentCompletion = null;
          clearCompletion(); // TODO: this is a bit aggressive?
        }
      }

      if (props.onKeyDown) {
        props.onKeyDown(event);
      }
    }
  }));
}
var EditorHelper = {
  isCompletionActive: function isCompletionActive(editor) {
    var completion = editor.children.find(function (n) {
      return n.type === 'completion';
    });
    return !!completion;
  },
  isCursorAtEndOfDocument: function isCursorAtEndOfDocument(editor) {
    var sel = editor.selection;
    if (!sel) {
      return false;
    }
    var _sel$anchor$path = sel.anchor.path,
      node = _sel$anchor$path[0];
    var offset = sel.anchor.offset;
    if (node !== editor.children.length - 1) {
      // not at the last node
      return false;
    }
    return offset === slate.Node.string(editor.children[node]).length;
  },
  acceptCompletion: function acceptCompletion(editor, fetcher, completion) {
    var isActive = EditorHelper.isCompletionActive(editor);
    if (isActive && completion != null && completion.completion) {
      EditorHelper.removeCompletion(editor);
      editor.insertText(completion.completion);
      fetcher.updateCompletion(completion.sid, {
        accepted: true
      });
    }
  },
  addCompletion: function addCompletion(editor, fetcher, completion) {
    var isActive = EditorHelper.isCompletionActive(editor);
    var isCursorEnd = EditorHelper.isCursorAtEndOfDocument(editor);
    if (!isActive && isCursorEnd && completion != null && completion.completion) {
      slate.Transforms.insertNodes(editor, {
        type: 'completion',
        children: [{
          text: completion.completion
        }]
      }, {
        at: [editor.children.length]
      });
      fetcher.updateCompletion(completion.sid, {
        seen: true
      });
    }
  },
  removeCompletion: function removeCompletion(editor) {
    var isActive = EditorHelper.isCompletionActive(editor);
    if (isActive) {
      slate.Transforms.removeNodes(editor, {
        at: [editor.children.length - 1]
      });
    }
  },
  updateCompletion: function updateCompletion(editor, newCompletion) {
    var isActive = EditorHelper.isCompletionActive(editor);
    if (isActive && newCompletion.length > 0) {
      // could not find a way to update to set the text of the node, so we remove the first character only
      slate.Transforms["delete"](editor, {
        at: {
          anchor: {
            path: [editor.children.length - 1, 0],
            offset: 0
          },
          focus: {
            path: [editor.children.length - 1, 0],
            offset: 1
          }
        }
      });
    }
  },
  serialize: function serialize(nodes) {
    return nodes.map(function (n) {
      return slate.Node.string(n);
    }).join('\n');
  }
};

// export * from './SmartInput'

module.exports = SmartInput;
//# sourceMappingURL=index.cjs.map
